{
  TAlLanguagesStore
  TAlLanguager

  ************
  (C) 1999 Aldo Ghigliano aldog@ivg.it aldo.ghigliano@iname.com
           Portions (C) RxLib, AO ROSNO, Master-Bank
  This is freeware, and "As Is", but:
  - if you modify this unit, you have to mail to me all modifications,
    and to obtain my acknowledgment before to distribute them;
  - gift (may be 5$, send to Ghigliano Aldo - Viale Riviera 82 -
    17027 Pietra Ligure - Italy) are well accepted;
  WARNING
  Aldo Ghigliano is not responsible of anything happens using this unit.
  Use it at your own risk

  ************
  PURPOSE:
      Localization (language translation) of forms for large projects
    by automatically using conditional compiler directives (ifdef..endif).
  PLATFORMS:
      Delphi3. Maybe 4.
  NEEDS:
      Rx Lib of MasterBank. You can download it from DSP.
  CREDITS:
      I have the idea after I have see TVCLlocalizer by
    Markus Stephany mailto:mirbir.st@t-online.de
  INSTALLATION:
      Do "Component >Install Component" of uAlLangs.pas, and insert in
    the package you wants, as usual.
  USAGE:
      To localize your project:
      1)Put an AlLanguagesStore to mainform of your project, and set
    LanguagesPasUnit='c:\yourProjectPath\YourLocFormsUnit.pas'.
    This file don't need to exists.
    Set also the DefaultLang (eng is the predefined), and put a list
    of the defines you will use in conditional compilation in
    LanguagesDefines property.
      2)Put an AlLanguager in any forms of your project.
      3)Add in the OnCreate event handler of every form, the code
    you can see in the demo.
      4)Last, link these AlLanguager to AlLanguagesStore.
      If you change something in a form, simply double-click on the
    AlLanguager to update strings in the default language.
      To translate, open 'c:\yourProjectPath\YourLocFormsUnit.pas',
    and simply edit.
      See directory demo for a simple example on how to set project
    options to conditional compile the project in the language you
    want. In the demo, projIta.dpr runs italian, projEng.dpr runs
    english. They share same units.
    Language is defined at compilation, thus:
      -executable size is lower
      -you can localize also other program messages by "ifdef"s
       (but use another '.pas' unit to avoid problems)
      -only one language is available at runtime
      Note you can use many AlLanguagesStore as you want; this is useful
    if for example your program has different versions (i.e. "Standard"
    edition "Educational",...) that shares some units, and some others not.


NOTE
 - this unit is not planned to be localized itself... to avoid
   confusion :)
 - let me know your suggestions / bug reports
 Enjoy.
 Al
}

unit uAlLangs;
interface
uses
  windows,
  SysUtils, Classes, Graphics, Controls, Forms, TypInfo,
  DsgnIntf, VclUtils;

type
  TAlLanguagesStore=class(TComponent)
  private
    FDefaultLang:string;
    FLanguagesDefines: TStrings;
    FLanguagesPasUnit: TFileName;
    procedure SetLanguagesDefines(value: TStrings);
    procedure SetLanguagesPasUnit(afileName: TfileName);
    //
    function  CreateFileIfNeeded:boolean;
    procedure CreateFile;
    procedure AddBlocksIfNeeded;
    procedure AddBlocksIfNeededSt(stream:TMemoryStream);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Update(AFormName:string;LocalizationStrings:TStrings);
    procedure Remove(AFormName:string;LocalizationProp:String);
  published
    property  DefaultLang:string read FDefaultLang write FDefaultLang;
    property  LanguagesPasUnit:TFileName read FLanguagesPasUnit write SetLanguagesPasUnit;
    property  LanguagesDefines:TStrings read FLanguagesDefines write SetLanguagesDefines;
  end;

  TAlLanguager=class(TComponent)
  private
    //owner is the owner Form, ever;
    FLanguagesStore:TAlLanguagesStore;
    FPropertiesList: TStrings;
    FLocalizationStrings: TStrings;
    FUpdated:boolean;
    //
    procedure SetLocalizationString(Value: string);
    procedure SetPropertiesList(Value: TStrings);
    procedure SetUpdate(Value: TStrings);
    function  GetPropertiesList:TStrings;
    function  GetUpdate:TStrings;
    //
    procedure DoWriteString(const Section, Ident, Value: string);
    function  DoReadString(const Section, Ident,
      aDefault: string): string;
    procedure DoEraseSection(const Section: string);
    //
    procedure AddPropOf(c:TComponent);
    function  PropertiesListFromForm:TStrings;
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {$IFDEF WIN32}
    procedure SetNotification;
    {$ENDIF WIN32}
  published
    property LanguagesStore:TAlLanguagesStore read FLanguagesStore write FLanguagesStore;
    property LocalizationString:string write SetLocalizationString stored false;
    property PropertiesList: TStrings read GetPropertiesList write SetPropertiesList stored false;
  end;

  TAlLanguagerEditor = class(TComponentEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

procedure Register;

implementation
uses
  menus,RxProps;
{$R *.dcr}
procedure Register;
begin
  RegisterComponents('Language',[TAlLanguager]);
  RegisterComponents('Language',[TAlLanguagesStore]);
  RegisterComponentEditor(TAlLanguager, TAlLanguagerEditor);
end;

{
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
  stuffs
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
}
const
  srLanguagerEditor='&Update strings';
  _StrStopLang:string= #13#10'{$endif}//AlLanguagesStore';
  _StrStopForm:string= #13#10'       ;//AlLanguagesStore';
  _LocStr:string=#13#10'    +'' ''+{##>}''';
  _LocStr2:string=#13#10'    +''';
  _LocStr3:string='''{<##}+#13#10 ';
  _Remark=#13#10'//STRING mantained by AlLanguagesStore.'
         +#13#10'//Please modify only strings marked by {##>} {<##}.';
  msg_m_info='Information';
  msg_m_warn='Warning!';

procedure Msg(s:string);
var ss:array[0..254]of char;
begin
  strPcopy(ss,s);
  messagebox(0,ss,msg_m_info,mb_ok+mb_iconinformation
    +mb_taskmodal);
end;

procedure Msgw(s:string);
var ss:array[0..254]of char;
begin
  strPcopy(ss,s);
  messagebox(0,ss,msg_m_warn,mb_ok+mb_iconexclamation
    +mb_taskmodal);
end;

//must speed up? don't know, but following may help.
//for other system you may replace with a classic StrPos
//cnt isn't really used.
function AlStrLPos(str,substr:pchar;cnt:integer):pchar; assembler;
asm // must preserve the EDI, ESI, ESP, EBP, and EBX
        push edi
        push esi
        push ebx
        mov esi,str
        mov ebx,esi
        mov ecx,cnt
        and ecx,ecx
        jz @found
@ricom: mov edi,substr
        mov ah,[edi]
        and ah,ah
        jz @found
@lop:   mov al,[esi]
        cmp al,ah
        jz @firstNextcar
        and al,al
        jz @endst
        inc esi
        dec ecx
        jnz @lop
        jmp @endst
@firstnextcar:
        mov ebx,esi
@nextcar:
        inc edi
        inc esi
        mov ah,[edi]
        and ah,ah
        jz @found
        cmp ah,[esi]
        jz @nextcar
        jmp @ricom
@endst: mov ebx,0
@found: //result in eax
        mov eax,ebx
        pop ebx
        pop esi
        pop edi
end;

function Spezza(value:string):string;
var
  l,ll:integer;
begin
  ll:=length(value)-1;
  if (ll<80) then begin
    result:=copy(value,2,maxint);
  end else begin
    result:='';
    for l:=0 to (ll div 80) do begin
      result:=result+copy(value,2+l*80,80)+#39#13#10'      +'#39;
    end;
  end;
end;








{
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
  TAlLanguagesStore
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
}
{
-------------------------------------------------------------------------
}
constructor TAlLanguagesStore.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLanguagesDefines:=TStringList.Create;
  FDefaultLang:='eng';
  FLanguagesDefines.add(FDefaultLang);
  FLanguagesPasUnit:='c:\FormLocalizations1.pas';
end;

destructor TAlLanguagesStore.Destroy;
begin
  FLanguagesDefines.Free;
  FLanguagesDefines:=NIL;
  inherited Destroy;
end;


{
-------------------------------------------------------------------------
}
//update .pas file with new strings to localize
procedure TAlLanguagesStore.Update(AFormName:string;LocalizationStrings:TStrings);
var
  stream:TMemoryStream;
  streamf:TFileStream;
  StrStartLang,StrStopLang,StrStartForm,StrStopForm,
  ToAddStrings,NewStrings,st,t,s:string;
  pc,pcc,startL,stopL,startF,stopF:pchar;
  kk,k,tomovelen:integer;
  ok:boolean;
label
  fine;
begin
  ok:=false;
  CreateFileIfNeeded;
  //msg('update');
  //updates default language, getting from form strings to
  //localize
  stream:=TMemoryStream.create;
  try
    stream.LoadFromFile(FLanguagesPasUnit);
    AddBlocksIfNeededSt(stream);
    pc:=pchar(stream.memory);
    //
    StrStartLang:=#13#10'{$ifdef '+FDefaultLang+'}';
    StrStartForm:=#13#10'  Lang_'+aformName+'='#39#39;
    NewStrings:=StrStartForm;
    for k:=0 to LocalizationStrings.count-1 do begin
      if (k mod 2)<>0 then begin
        NewStrings:=newstrings
          +_LocStr
          +Spezza(LocalizationStrings[k])+_LocStr3;
      end else begin
        NewStrings:=newstrings
          +_LocStr2+LocalizationStrings[k]+'''#13#10';
      end;
    end;
    //msg('update 1');
    NewStrings:=NewStrings+_StrStopForm;
    pcc:=AlStrLPos(pc,pchar(StrStartLang),maxint);
    //
    if pcc=NIL then goto fine;
    startL:=pcc+length(StrStartLang)+2;
    stopL:=AlStrLPos(pcc,pchar(_StrStopLang),maxint);
    pcc:=AlStrLPos(startL,pchar(StrStartForm),maxint);
    if (pcc<>NIL) and (pcc<stopL) then begin
      //strings relative to present form already present
      startF:=pcc;
      stopF:=AlStrLPos(startF,pchar(_StrStopForm),stopL-startF)
        +length(_StrStopForm);
      if (stopF=NIL) or (stopF>stopL) then goto fine; //out of present form
      tomovelen:=stream.size-(stopF-pc);
      setLength(st,tomovelen);
      stream.position:=stopF-pc;
      stream.readbuffer(pchar(st)^,tomovelen);
      stream.position:=startF-pc;
      stream.writeBuffer(pchar(NewStrings)^,length(NewStrings));
      stream.writeBuffer(pchar(st)^,tomovelen);
      stream.size:=stream.position;
      //msg('update 2');
    end else begin
      //present form have not his own string in this language.
      //I insert it now
      tomovelen:=stream.size-(stopL-pc);
      setLength(st,tomovelen);
      stream.position:=stopL-pc;
      stream.readbuffer(pchar(st)^,tomovelen);
      stream.position:=stopL-pc;
      stream.writeBuffer(pchar(NewStrings)^,length(NewStrings));
      stream.writeBuffer(pchar(st)^,tomovelen);
      stream.size:=stream.position;
      //msg('update 3');
    end;
    pc:=pchar(stream.memory);
    //Scans other languages ifdef..endif.
    //If some couple property-value isn't present, I add it
    //copying from defaultLang one.
    //User must edit LanguagesPasUnit to do effective localization.
    for k:=0 to FLanguagesDefines.count-1 do begin
      //skips def lang
      if FLanguagesDefines[k]=FDefaultLang then continue;
      //search for current lang start and stop (ifdef..endif)
      StrStartLang:=#13#10'{$ifdef '+FLanguagesDefines[k]+'}';
      pcc:=AlStrLPos(pc,pchar(StrStartLang),maxint);
      startL:=pcc+length(StrStartLang)+2;
      stopL:=AlStrLPos(startL,pchar(_StrStopLang),maxint);
      //search for current form start and stop (Lang_form='...';)
      pcc:=AlStrLPos(startL,pchar(StrStartForm),stopL-startL);
      if (pcc<>NIL) and (pcc<stopL) then begin
        startF:=pcc;
        stopF:=AlStrLPos(startF,pchar(_StrStopForm),stopL-startF);
        if (stopF=NIL) or (stopF>stopL) then goto fine;
        //current form have his own string yet, so I must
        //scan the string to add property in NewStrings,
        //but not in (Lang_form='...';).
        ToAddStrings:='';
        for kk:=0 to (LocalizationStrings.count div 2)-1 do begin
          s:=_LocStr2+LocalizationStrings[kk*2]+'''#13#10';
          pcc:=AlStrLPos(startF,pchar(s),stopL-startF);
          if (pcc=NIL) or (pcc>stopL) then begin
            //there isn't a couple property-value: I'll add it
            ToAddStrings:=ToAddStrings+s
              +_LocStr
              +Spezza(LocalizationStrings[kk*2+1])+_LocStr3;
          end;
        end;
        if ToAddStrings<>'' then begin
          tomovelen:=stream.size-(stopF-pc);
          setLength(st,tomovelen);
          stream.position:=stopF-pc;
          stream.readbuffer(pchar(st)^,tomovelen);
          stream.position:=stopF-pc;
          stream.writeBuffer(pchar(ToAddStrings)^,length(ToAddStrings));
          stream.writeBuffer(pchar(st)^,tomovelen);
          stream.size:=stream.position;
          //msg('Added:'#10+ToAddStrings);
        end;
      end else begin
        //current form doesn't have his own string yet, so I must
        //insert entire string, copying from default lang
        tomovelen:=stream.size-(stopL-pc);
        setLength(st,tomovelen);
        stream.position:=stopL-pc;
        stream.readbuffer(pchar(st)^,tomovelen);
        stream.position:=stopL-pc;
        stream.writeBuffer(pchar(NewStrings)^,length(NewStrings));
        stream.writeBuffer(pchar(st)^,tomovelen);
        stream.size:=stream.position;
      end;
      pc:=pchar(stream.memory);
    end;
    //
    stream.SaveToFile(FLanguagesPasUnit);
    ok:=true;
    //
    fine:
    if not ok then begin
      msgw('Something wrong updating strings of: '+AFormName);
    end;
  finally
    if stream<>NIL then stream.free;
  end;
end;

//remove a property from all langs (not used and tested yet)
procedure TAlLanguagesStore.Remove(AFormName:string;LocalizationProp:String);
var
  stream:TMemoryStream;
  streamf:TFileStream;
  StrProp,
  StrStartLang,StrStopLang,StrStartForm,StrStopForm,
  ToAddStrings,NewStrings,st,t,s:string;
  pc,pcc,propStart,propStop,startL,stopL,startF,stopF:pchar;
  kk,k,tomovelen:integer;
  ok:boolean;
label
  fine;
begin
  ok:=false;
  if CreateFileIfNeeded then exit;
  stream:=TMemoryStream.create;
  try
    stream.LoadFromFile(FLanguagesPasUnit);
    AddBlocksIfNeededSt(stream);
    pc:=pchar(stream.memory);
    //
    //Scanno i linguaggi: se in essi esiste la prop, la tolgo.
    for k:=0 to FLanguagesDefines.count-1 do begin
      //cerco inizio e fine corrente linguaggio
      StrStartLang:=#13#10'{$ifdef '+FLanguagesDefines[k]+'}';
      pcc:=AlStrLPos(pc,pchar(StrStartLang),maxint);
      startL:=pcc+length(StrStartLang)+2;
      stopL:=AlStrLPos(startL,pchar(_StrStopLang),maxint);
      //cerco inizio e fine corrente form nel corrente linguaggio
      pcc:=AlStrLPos(startL,pchar(StrStartForm),stopL-startL);
      if (pcc<>NIL) and (pcc<stopL) then begin
        startF:=pcc;
        stopF:=AlStrLPos(startF,pchar(_StrStopForm),stopL-startF);
        if (stopF=NIL) or (stopF>stopL) then continue;
        //c'e' gi una stringa relativa al form in corso di update,
        //quindi devo passarla linea per linea,
        //e togliere le linee che competono alla prop LocalizationProp
        //e sono tra [startF e stopF].
        StrProp:=_LocStr2+LocalizationProp+'''#13#10';
        propStart:=AlStrLPos(startF,pchar(StrProp),stopF-startF);
        if (propStart=NIL) or (propStart>stopF) then continue;
        propStop:=AlStrLPos(propStart,pchar(_LocStr3),stopF-propStart);
        if (propStop=NIL) or (propStop>stopF) then continue;
        propStop:=propStop+length(_locStr3);
        //elimino tra propStart e propStop
        tomovelen:=stream.size-(propStop-pc);
        setLength(st,tomovelen);
        stream.position:=propStop-pc;
        stream.readbuffer(pchar(st)^,tomovelen);
        stream.position:=propStart-pc;
        stream.writeBuffer(pchar(st)^,tomovelen);
        stream.size:=stream.position;
      end;
      pc:=pchar(stream.memory);
    end;
    //
    stream.SaveToFile(FLanguagesPasUnit);
    ok:=true;
    //
    fine:
    if not ok then begin
      msgw('Something wrong deleting a couple property-value in: '+AFormName);
    end;
  finally
    if stream<>NIL then stream.free;
  end;
end;

procedure TAlLanguagesStore.SetLanguagesDefines(value: TStrings);
begin
  FLanguagesDefines.assign(value);
  if FLanguagesDefines.count=0 then FLanguagesDefines.add(FDefaultLang);
  if not CreateFileIfNeeded then
    AddBlocksIfNeeded;
  //add or decrease languages.
  //If I add, I'll create new language strings copying from default
  //lang.
  //If I delete, don't do anything (else user can't undo if I del!!!)
end;

procedure TAlLanguagesStore.SetLanguagesPasUnit(afileName: TfileName);
begin
  FLanguagesPasUnit:=afileName;
  //change file name
end;

procedure TAlLanguagesStore.AddBlocksIfNeeded;
var
  stream:TMemoryStream;
begin
  //msg('add block');
  stream:=TMemoryStream.create;
  stream.LoadFromFile(FLanguagesPasUnit);
  AddBlocksIfNeededSt(stream);
  stream.SaveToFile(FLanguagesPasUnit);
end;

procedure TAlLanguagesStore.AddBlocksIfNeededSt(stream:TMemoryStream);
var
  streamf:TFileStream;
  firstLang,st,t,block:string;
  pc,pcc,pcd,start,stop:pchar;
  ToAdd:TStringList;
  k,tomovelen:integer;
begin
  ToAdd:=TStringList.create;
  try
    pc:=pchar(stream.memory);
    firstLang:='';
    pcc:=AlStrLPos(pc,#13#10'const',maxint);
    pcd:=AlStrLPos(pc,#13#10'implementation',maxint);
    if (pcc=NIL) or (pcd=NIL) then begin
      CreateFile;
      stream.size:=0;
      stream.LoadFromFile(FLanguagesPasUnit);
    end;
    pc:=pchar(stream.memory);
    //
    for k:=0 to FLanguagesDefines.count-1 do begin
      block:=#13#10'{$ifdef '+FLanguagesDefines[k]+'}';
      //msg('blok');
      pcc:=AlStrLPos(pc,pchar(block),maxint);
      if pcc=NIL then begin
        ToAdd.add(FLanguagesDefines[k])
      end else if FirstLang='' then begin
        start:=pcc+length(block)+2;
        stop:=AlStrLPos(pcc,pchar(_StrStopLang),maxint);
        setLength(FirstLang,stop-start);
        stream.position:=start-pc;
        stream.readbuffer(pchar(FirstLang)^,stop-start);
      end;
    end;
    t:='';
    if ToAdd.count>0 then begin
      for k:=0 to ToAdd.count-1 do begin
        t:=t
        +_Remark
        +#13#10'{$ifdef '+ToAdd[k]+'}'
        +#13#10+FirstLang
        +_StrStopLang
        +#13#10''
        +#13#10'';
      end;
      pcc:=AlStrLPos(pc,#13#10'implementation',maxint);
      if pcc<>NIL then begin
        tomovelen:=stream.size-(pcc-pc);
        setLength(st,tomovelen);
        stream.position:=pcc-pc;
        stream.readbuffer(pchar(st)^,tomovelen);
        stream.position:=pcc-pc;
        stream.writeBuffer(pchar(t)^,length(t));
        stream.writeBuffer(pchar(st)^,tomovelen);
      end;
    end;
  finally
    ToAdd.free;
  end;
end;

function TAlLanguagesStore.CreateFileIfNeeded:boolean;
begin
  //se cambio file, creo il file?
  result:=not fileexists(FLanguagesPasUnit);
  if result then begin
    CreateFile;
  end else begin
    //s:=TFileStream.create(FLanguagesPasUnit,fmOpenRead or fmShareDenyWrite);
  end;
end;

procedure TAlLanguagesStore.CreateFile;
var
  s:TFileStream;
  t:string;
  k:integer;
begin
  s:=NIL;
  try
    s:=TFileStream.create(FLanguagesPasUnit,fmCreate or fmShareDenyWrite);
    t:='unit '+changeFileExt(extractFileName(FLanguagesPasUnit),'')+';'
      +#13#10'interface'
      +#13#10
      +#13#10'const'
      +#13#10;
    for k:=0 to FLanguagesDefines.count-1 do begin
      t:=t
      +_Remark
      +#13#10'{$ifdef '+FLanguagesDefines[k]+'}'
      +#13#10''
      +_StrStopLang
      +#13#10''
      +#13#10'';
    end;
    t:=t
      +#13#10'implementation'
      +#13#10'end.';
    s.writeBuffer(pchar(t)^,length(t));
  finally
    if s<>NIL then s.free;
  end;
end;






{
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
  TAlLanguager
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
}
{
-------------------------------------------------------------------------
}
constructor TAlLanguager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FUpdated:=false;
  FLanguagesStore:=NIL;
  FPropertiesList:=TStringList.Create;
  FLocalizationStrings:= TStringList.Create;
end;

destructor TAlLanguager.Destroy;
begin
  FLocalizationStrings.Free;
  FLocalizationStrings:=NIL;
  FPropertiesList.Free;
  FPropertiesList:=NIL;
  inherited Destroy;
end;


{
-------------------------------------------------------------------------
}
const
  tkLangString=[tkString,tkClass,tkLString,tkWString];

procedure TAlLanguager.AddPropOf(c:TComponent);
var
  list: TPropInfoList;
  j:    integer;
  pnam: string;
begin
  if c.name='' then exit;
  list := TPropInfoList.Create(C, tkLangString);
  try
    try
      for j := 0 to list.Count - 1 do begin
            pnam:=lowercase(list.Items[j]^.Name);
            if (c is TMenuItem) or (c is TCustomForm) then begin
              if (pnam='caption') or (pnam='text') then
                FPropertiesList.AddObject(
                  lowercase(c.name)+'.'+pnam, c);
            end else begin
              if (pnam='caption') or (pnam='text') or (pnam='hint')
              or (pnam='strings') or (pnam='items') or (pnam='lines') then
                FPropertiesList.AddObject(
                  lowercase(c.name)+'.'+pnam, c);
            end;
      end;
    except
    end;
  finally
    list.free;
  end;
end;

//-------------------
//Scans the form; gets a list of all localizable properties
//(captions, hints, items, ...)
function TAlLanguager.PropertiesListFromForm:TStrings;
var
  i: Integer;
  c: TComponent;
  fOwner:TComponent;
  pnam:string;
begin
  result:=FPropertiesList;
  if (csDestroying in ComponentState) then exit;
  //msg('Reading...');
  FPropertiesList.clear;
  fOwner:=owner;
  AddPropOf(fowner); //caption del form
  if FOwner <> nil then begin
    for i := 0 to FOwner.ComponentCount - 1 do begin
      c := FOwner.Components[I];
      if not (c is TAlLanguager) then AddPropOf(c);
    end;
  end;
  //i.e.: in a form with 2 label, you will have on exit:
  //  FCompToLangString.count=4;
  //  FCompToLangString.items
  //    label1.caption
  //    label1.hint
  //    label2.caption
  //    label2.hint
end;

//when the form was loaded, we can read the list of localizable
//properties
procedure TAlLanguager.Loaded;
begin
  inherited Loaded;
  //PropertiesListFromForm;
  //UpdateStoredList(Owner, PropertiesListFromForm, True);
end;




{
-------------------------------------------------------------------------
}
{$IFDEF WIN32}
procedure TAlLanguager.SetNotification;
var
  i: Integer;
  Component: TComponent;
begin
  exit;
  for i := FPropertiesList.Count - 1 downto 0 do begin
    Component := TComponent(FPropertiesList.Objects[i]);
    if Component <> nil then Component.FreeNotification(Self);
  end;
end;
{$ENDIF WIN32}

procedure TAlLanguager.Notification(AComponent: TComponent; Operation: TOperation);
var
  I: Integer;
  Component: TComponent;
begin
  inherited Notification(AComponent, Operation);
  if (FPropertiesList = nil) then exit;
  if not (csDestroying in ComponentState) then begin
    //
    if AComponent is TAlLanguager then exit;
    //
    if AComponent is TAlLanguagesStore then begin
      if AComponent=FLanguagesStore then FLanguagesStore:=NIL;
      exit;
    end;
    //
    //probabilly following does nothing
    if csDesigning in AComponent.ComponentState then
    case Operation of
      opInsert:begin
        FUpdated:=false;
        //msg('insert');
        //AddPropOf(acomponent);
      end;
      opRemove:begin
        for I := FPropertiesList.Count - 1 downto 0 do begin
          Component := TComponent(FPropertiesList.Objects[I]);
          if Component = AComponent then begin
            {
            if FAlLanguagesStore<>NIL then
              FAlLanguagesStore.Remove(
                owner.Name,'*_*'+FPropertiesList[i]);
            }
            FPropertiesList.Delete(I);
            //e aggiorno il FLocalStringFile
          end;
        end;
        //msg('remove');
      end;
    end;
  end;
end;





//-----------------------------
procedure TAlLanguager.SetPropertiesList(Value: TStrings);
begin
  //
end;

procedure TAlLanguager.SetUpdate(Value: TStrings);
begin
  //
end;

procedure TAlLanguager.SetLocalizationString(Value: string);
begin
  if not FUpdated then PropertiesListFromForm;
  FLocalizationStrings.text:=Value;
  {$IFDEF WIN32}
  SetNotification;
  {$ENDIF}
  with TPropsStorage.Create do
  try
    Section := '_';//IniSection;
    OnReadString := DoReadString;
    try
      LoadObjectsProps(Owner, FPropertiesList);
    except        // ignore any exceptions
    end;
  finally
    Free;
  end;
  //msg('loaded loc str from assignment');
end;

//------------------
function TAlLanguager.GetPropertiesList: TStrings;
begin
  result:=FPropertiesList;
  if (FPropertiesList=NIL) or
     (csDestroying in ComponentState) then exit;
  if not FUpdated then begin
    FUpdated:=true;
    PropertiesListFromForm;
  end;
end;

//-----------------------
//Works design-time only. When user clicks "Update", here
//we read values of properties listed in PropertiesList,
//then these values are saves either in TAlLanguagesStore
//and  FLocalizationStrings
function TAlLanguager.GetUpdate:TStrings;
var
  s:string;
begin
  result:=FLocalizationStrings;
  if not FUpdated then begin
    FUpdated:=true;
    PropertiesListFromForm;
  end;
  s:='Properties values updated from form.';
  if FLanguagesStore<>NIL then s:=s+#10'Pas file: '
    +FLanguagesStore.LanguagesPasUnit;
  s:='Remember you should update other forms separately.';
  msg(s);
  if not (csDesigning in componentState) then exit;
  FLocalizationStrings.clear;
  with TPropsStorage.Create do
  try
    Section := '_';//IniSection;
    OnWriteString := DoWriteString;
    OnEraseSection:= DoEraseSection;
    StoreObjectsProps(Owner, FPropertiesList);
  finally
    Free;
  end;
  if FLanguagesStore<>NIL then begin
    FLanguagesStore.update(owner.name,FLocalizationStrings);
  end else
    msgw('AlLanguagesStore undefined!');
end;




//----------------------
function ElReturn(value:string):string;
var
  l,ll:integer;
  pc:pchar;
begin
  result:='';
  l:=1;
  ll:=length(value)+1;
  while l<ll do begin
    case value[l] of
      #13:if (value[l+1]=#10) then begin
        inc(l,2);
        result:=result+'!';
      end else begin
        inc(l);
        result:=result+'!';
      end;
      #10:begin
        inc(l);
        result:=result+'!';
      end;
      '':begin
        inc(l);
        result:=result+'';
      end;
      #39:begin //'
        inc(l);
        result:=result+#39#39;
      end;
      else begin
        result:=result+value[l];
        inc(l);
      end;
    end;
  end;
end;

function ElDeReturn(value:string):string;
var
  l,ll:integer;
  pc:pchar;
begin
  result:='';
  l:=1;
  ll:=length(value)+1;
  while l<ll do begin
    case value[l] of
      '':if (value[l+1]='!') then begin
        inc(l,2);
        result:=result+#13#10;
      end else if (value[l+1]='') then begin
        inc(l,2);
        result:=result+'';
      end else inc(l);
      #39:if (value[l+1]=#39) then begin
        inc(l,2);
        result:=result+#39;
      end else begin
        result:=result+#39;
        inc(l);
      end;
      else begin
        result:=result+value[l];
        inc(l);
      end;
    end;
  end;
end;

procedure TAlLanguager.DoWriteString(const Section, Ident, Value: string);
begin
  if value<>'' then begin
    FLocalizationStrings.add('*'+Section+'*'+Ident);
    FLocalizationStrings.add(' '+ElReturn(value));
  end;
end;

function TAlLanguager.DoReadString(const Section, Ident,
           aDefault: string): string;
var
  i:integer;
begin
  i:=FLocalizationStrings.IndexOf('*'+lowercase(section)+'*'+lowercase(ident));
  if (i<>-1) and (FLocalizationStrings.count>=i) then begin
    result:=
      ElDeReturn(
        copy(FLocalizationStrings[i+1],2,maxint) );
  end else result:=adefault;
end;

procedure TAlLanguager.DoEraseSection(const Section: string);
begin
  //parto sempre da FcurLangString nullo.
end;








{
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
  TAlLanguagerEditor
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
}
procedure TAlLanguagerEditor.ExecuteVerb(Index: Integer);
var
  al: TAlLanguager;
begin
  al := Component as TAlLanguager;
  if Index = 0 then begin
    //msg('Verb');
    al.getUpdate;
  end;
end;

function TAlLanguagerEditor.GetVerb(Index: Integer): string;
begin
  case Index of
    0: Result := srLanguagerEditor;
    else Result := '';
  end;
end;

function TAlLanguagerEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;





end.


